/*
 * SIP extension for IP connection tracking. 
 */

#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/tcp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_sip.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>

#if defined(VOICE_QUALITY_CONTROL_BY_ATM)
#include <bcmtypes.h>
#include <bcmatmapi.h>
#endif

DECLARE_LOCK(ip_sip_lock);
 
struct module *ip_conntrack_sip = THIS_MODULE;

#define MAX_PORTS 8
static int ports[MAX_PORTS];
static int ports_c;

#if defined(VOICE_QUALITY_CONTROL_BY_ATM)
static unsigned int stunaddr=0;
static unsigned int stunmask=0;
static int stunaddr_first;
static int stunaddr_end;
static int stunmask_first;
static int stunmask_end;
static int vpi;
static int vci;
static int protocol;
static int encapmode;
module_param(stunaddr_first, int,0);
module_param(stunaddr_end, int,0);
module_param(stunmask_first, int,0);
module_param(stunmask_end, int,0);
module_param(vpi, int, 0);
module_param(vci, int, 0);
module_param(protocol, int, 0);//1:pppoe, 2:pppoa, 3:mer, 4:ipoa, 5:bridge
module_param(encapmode, int, 0);//0:LLC/SNAP, 1:VCMUX (PPPOA 0:VCMUX, 1:LLC/SNAP)
#else
#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif
#endif

#if 0
  #define DEBUGP printk
#else
  #define DEBUGP(format, args...)
#endif

#define RTCP_SUPPORT

struct ip_conntrack_expect *
expect_find(struct ip_conntrack *ct, u_int32_t dstip, u_int16_t dstport)
{
	const struct ip_conntrack_tuple tuple = ((struct ip_conntrack_tuple)
		{ { 0, { 0 } }, { dstip, { .udp = { htons(dstport) } }, IPPROTO_UDP }});

    return ip_conntrack_expect_find_get(&tuple);
}

static int set_expected_voip(struct ip_conntrack *ct, u_int32_t dstip,
		u_int16_t dstport, unsigned int conntype)
{
	int ret = 0;
	struct ip_ct_sip_expect *exp_sip_info;
	struct ip_conntrack_expect expect, *exp = NULL;

	memset(&expect, 0, sizeof(struct ip_conntrack_expect));

	LOCK_BH(&ip_sip_lock);

	DEBUGP("conntrack_sip: %s: [%s]: DST=%u.%u.%u.%u:%u\n", __FUNCTION__, 
			((conntype == CONN_SIP) ? "SIP" 
									: ((conntype == CONN_RTP) ? "RTP" : "RTCP")), 
			NIPQUAD(dstip), dstport);

	/* exp_sip_info */
	exp_sip_info = &expect.help.exp_sip_info;
	exp_sip_info->port = dstport;
	exp_sip_info->type = conntype;
	exp_sip_info->nated = 0;

	/* new expectation */
	expect.tuple = ((struct ip_conntrack_tuple)
		{ { 0, { 0 } },
		  { dstip, { .udp = { htons(dstport) } }, IPPROTO_UDP }});

    expect.mask = ((struct ip_conntrack_tuple)
        { { 0, { 0 } },
          { 0xFFFFFFFF, { .all = {0xFFFFFFFFFFFFFFFF} }, 0xFF }});
	//memset(&expect.mask, 0xFF, sizeof(struct ip_conntrack_tuple));

	expect.expectfn = NULL;

	exp = ip_conntrack_expect_alloc();
	if (exp == NULL) return NF_ACCEPT;
	memcpy(exp, &expect, sizeof(struct ip_conntrack_expect));

	if ((ret=ip_conntrack_expect_related(exp, ct)) != 0) {
		DEBUGP("  add expectation fail.. with exp=[%p], conntype=[%d], "
				"ret=[%d], ip=%u.%u.%u.%u, port=%d\n", 
				exp, conntype, ret, NIPQUAD(dstip), 
				ntohs(exp->tuple.dst.u.udp.port));
		if(conntype == CONN_SIP && ret == -EEXIST) {
			struct ip_conntrack_expect *expo = expect_find(ct, dstip, dstport);
#ifdef DEBUG_PRINTK
			if(expo == NULL)
				DEBUGP("find expectation for dstip=%u.%u.%u.%u port=%d fail.\n", 
						NIPQUAD(dstip), ntohs(exp->tuple.dst.u.udp.port));
#endif
			if (expo && ct != expo->expectant) {
				ip_conntrack_unexpect_related(expo);

				/*******************************************************************
				 * in kernel 2.6.x, if ip_conntrack_expect_related() return fail,  *
				 * it will free the memory which exp point. we need to re-allocate *
				 * new memory for ip_conntrack_expect_related() again.             *
				 *                               Andrew@odm.com.tw (2005/06/22)  *
				 *******************************************************************/
				exp = ip_conntrack_expect_alloc();
				if (exp == NULL) return NF_ACCEPT;
				memcpy(exp, &expect, sizeof(struct ip_conntrack_expect));

				if ((ret=ip_conntrack_expect_related(exp, ct)) != 0) {
					DEBUGP("  add expectation fail.. with exp=[%p], conntype=[%d], "
							"ret=[%d], ip=%u.%u.%u.%u, port=%d\n", 
							exp, conntype, ret, NIPQUAD(dstip), 
							ntohs(exp->tuple.dst.u.udp.port));
				} else {
					DEBUGP("  add expectation ok.. with exp=[%p], conntype=[%d], "
							"ret=[%d], ip=%u.%u.%u.%u, port=%d\n", 
							exp, conntype, ret, NIPQUAD(dstip), 
							ntohs(exp->tuple.dst.u.udp.port));
				}
			}
		} else {
			DEBUGP("Can't add new expectation for conntype=%d. \n", conntype);
		}
	} else {
		DEBUGP("  add expectation ok.. with exp=[%p], conntype=[%d], "
				"ret=[%d], ip=%u.%u.%u.%u, port=%d\n", 
				exp, conntype, ret, NIPQUAD(dstip), 
				ntohs(exp->tuple.dst.u.udp.port));
	}

	UNLOCK_BH(&ip_sip_lock);

	return ret;
}

/*
static int unset_expected_voip(struct ip_conntrack *ct, u_int32_t dstip, u_int16_t dstport)
{
	struct ip_conntrack_expect *exp;
	const struct ip_conntrack_tuple tuple = ((struct ip_conntrack_tuple)
		{ { 0, { 0 } }, { dstip, { .udp = { htons(dstport) } }, IPPROTO_UDP }});

	LOCK_BH(&ip_sip_lock);

	exp = ip_conntrack_expect_find_get(&tuple);
	if (exp) {
		DEBUGP("Find the expectation %p, then delete it.\n", exp);
		ip_conntrack_unexpect_related(exp);
	}

	UNLOCK_BH(&ip_sip_lock);

	return 0;
}*/

/* return:
 *       0 : Not found
 *       1 : Found domain name
 *       2 : Found dotted quads
 */
int find_sdp_rtp_addr(const char *data, size_t dlen,
			unsigned int *numoff, unsigned int *numlen, u_int32_t *addr)
{
	char *st, *p = (char *)data;
	const char *limit = data + dlen;
	unsigned char p1, p2, p3, p4;

	*numoff = 0;
	*numlen = 0;

	while (p < limit) {
		/* find 'c=' line */
		if (strncmp(p, "\nc=",3) && strncmp(p, "\rc=",3)) {
			p++;
			continue;
		}
		p += 3;

		if (strncmp(p, "IN IP4 ",7))
			continue;
		p += 7;

		/* IP address */
		st = p;

		/* FQDNs or dotted quads */
		while (isalpha(*p) || isdigit(*p) || (*p=='.') || (*p=='-')) {
			p++;
			if (p == limit)
				return 0;
		}

		*numoff = st - data;
		*numlen = p - st;

		/* Convert the IP address */
		p1 = simple_strtoul(st, &st ,10);
		if (*st != '.')
			return 1;
		p2 = simple_strtoul(st+1, &st, 10);
		if (*st != '.')
			return 1;
		p3 = simple_strtoul(st+1, &st, 10);
		if (*st != '.')
			return 1;
		p4 = simple_strtoul(st+1, &st, 10);

		*addr = (p1<<24) | (p2<<16) | (p3<<8) | p4;

		return 2;
	}

	return 0;
}

u_int16_t find_sdp_audio_port(const char *data, size_t dlen,
			unsigned int *numoff, unsigned int *numlen)
{
	char *st, *p = (char *)data;
	const char *limit = data + dlen;
	u_int16_t port = 0;

	*numoff = 0;
	*numlen = 0;

	while (p < limit) {
		/* find 'm=' */
		if (strncmp(p, "\nm=", 3) && strncmp(p, "\rm=", 3)) {
			p++;
			continue;
		}
		p += 3;

		/* audio stream */
		if (strncmp(p ,"audio ",6))
			continue;
		p += 6;

		st = p;
		port = simple_strtoul(p, &p, 10);

		*numoff = st - data;
		*numlen = p - st;

		return port;
	}

	return 0;
}

u_int16_t find_sdp_video_port(const char *data, size_t dlen,
			unsigned int *numoff, unsigned int *numlen)
{
	char *st, *p = (char *)data;
	const char *limit = data + dlen;
	u_int16_t port = 0;

	*numoff = 0;
	*numlen = 0;

	while (p < limit) {
		/* find 'm=' */
		if (strncmp(p, "\nm=", 3) && strncmp(p, "\rm=", 3)) {
			p++;
			continue;
		}
		p += 3;

		/* video stream */
		if (strncmp(p ,"video ",6))
			continue;
		p += 6;

		st = p;
		port = simple_strtoul(p, &p, 10);

		*numoff = st - data;
		*numlen = p - st;

		return port;
	}

	return 0;
}

u_int16_t find_sdp_image_port(const char *data, size_t dlen,
			unsigned int *numoff, unsigned int *numlen)
{
	char *st, *p = (char *)data;
	const char *limit = data + dlen;
	u_int16_t port = 0;

	*numoff = 0;
	*numlen = 0;

	while (p < limit) {
		/* find 'm=' */
		if (strncmp(p, "\nm=", 3) && strncmp(p, "\rm=", 3)) {
			p++;
			continue;
		}
		p += 3;

		/* T.38 image stream */
		if (strncmp(p ,"image ",6))
			continue;
		p += 6;

		st = p;
		port = simple_strtoul(p, &p, 10);

		*numoff = st - data;
		*numlen = p - st;

		return port;
	}

	return 0;
}

static int help(struct sk_buff *skb,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen, matchoff;
	struct iphdr *iph = skb->nh.iph;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)udph + 8;
	unsigned int udplen = skb->len - iph->ihl*4;
	unsigned int datalen = udplen - 8;

	u_int32_t ipaddr=0;
	u_int16_t port = 0;
	u_int16_t videoport = 0;//Andy add
	
	int found = 0;

	struct ip_ct_sip_master *ct_sip_info = &ct->help.ct_sip_info;

	DEBUGP("\nconntrack_sip: help(): DIR=%d, conntrackinfo=%u\n", dir, ctinfo);
	DEBUGP("conntrack_sip: %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n", 
		        NIPQUAD(ct->tuplehash[dir].tuple.src.ip),
		        ntohs(ct->tuplehash[dir].tuple.src.u.udp.port),
		        NIPQUAD(ct->tuplehash[dir].tuple.dst.ip), 
		        ntohs(ct->tuplehash[dir].tuple.dst.u.udp.port) );
	/* Reset for a new incoming packet */
	ct_sip_info->mangled = 0;	

	/* keep the connection alive */
	ip_ct_refresh(ct, (SIP_EXPIRES * HZ));

	/* Don't need to set the expectation for upstream direction */
	if (dir == IP_CT_DIR_REPLY)
		return NF_ACCEPT;

	/* Need to set the expected connection for further incoming RTP stream */
	if (strncmp(data, "INVITE", 6) == 0 || strncmp(data, "SIP/2.0 200", 11) == 0) {
		/* Find RTP address */
		found = find_sdp_rtp_addr(data, datalen, &matchoff, &matchlen, &ipaddr);
		if (!found){//may be is BEY respones message
#if defined(VOICE_QUALITY_CONTROL_BY_ATM)
			if (strstr(data,"BYE")){
				if (vci != 0 && protocol != 0 && ct_sip_info->rtpport_priogroup > 0){
					BcmAtm_SetPriorityPacketGroup(ct_sip_info->rtpport_priogroup,NULL,0);
					ct_sip_info->groupadded = 0;
					ct_sip_info->rtpport = 0;
				}
			}
#endif			
			return NF_ACCEPT;
		}	

		DEBUGP("conntrack_sip: 'IN IP4' is %s.\n", (found == 1) ? "FQDNs" : "dotted quads");

		/* If it's a null address, then the call is on hold */
		if (found == 2 && ipaddr == 0) {
			DEBUGP("conntrack_sip: Null address is found.\n");
			return NF_ACCEPT;
		}

		/* Find audio port, and we don't like the well-known ports,
		 * which is less than 1024 */
		port = find_sdp_audio_port(data, datalen, &matchoff, &matchlen);
		if (port < 1024)
			return NF_ACCEPT;

		DEBUGP("conntrack_sip: audio port=%d.\n", port);

		ipaddr = ct->tuplehash[dir].tuple.src.ip; 
		ct_sip_info->rtpport = port;

		/* RFC1889 - RTP uses an even port number and the corresponding RTCP
		 * stream uses the next higher (odd) port number. */
		if (set_expected_voip(ct, ipaddr, port, CONN_RTP) == 0) {
#ifdef RTCP_SUPPORT
			set_expected_voip(ct, ipaddr, port + 1, CONN_RTCP);
#endif
		}
		
		/*Start Andy add for vedio steram*/
		videoport = find_sdp_video_port(data, datalen, &matchoff, &matchlen);
		if (videoport < 1024)
			return NF_ACCEPT;
		if (videoport != port){//Video port != Audio port, setup new expection for video stream
			ipaddr = ct->tuplehash[dir].tuple.src.ip; 
			ct_sip_info->rtpport = videoport;	
			DEBUGP("conntrack_sip: vidio port=%d.\n", videoport);
			if (set_expected_voip(ct, ipaddr, videoport, CONN_RTPVIDEO) == 0) {
#ifdef RTCP_SUPPORT
				set_expected_voip(ct, ipaddr, videoport + 1, CONN_RTCP);
#endif
			}
		}		
		/*End Andy add for vedio steram*/
	} else if (strncmp(data, "REGISTER", 8) == 0  || strncmp(data, "SUBSCRIBE", 9) == 0) {
		ipaddr = ct->tuplehash[dir].tuple.src.ip;
		port = SIP_PORT;

		set_expected_voip(ct, ipaddr, port, CONN_SIP);
		return NF_ACCEPT;
	}
#if defined(VOICE_QUALITY_CONTROL_BY_ATM)	 
	else if (strncmp(data, "BYE", 3) == 0 || strncmp(data, "CANCEL", 6) == 0) {
		if (vci != 0 && protocol != 0 && ct_sip_info->rtpport_priogroup > 0){
			BcmAtm_SetPriorityPacketGroup(ct_sip_info->rtpport_priogroup,NULL,0);
			ct_sip_info->groupadded =0;
			ct_sip_info->rtpport = 0;
		}
		return NF_ACCEPT;
	} 
#endif	
	else {
		DEBUGP("conntrack_sip: Not interesting packet.\n");
		return NF_ACCEPT;
	}


	return NF_ACCEPT;
}

static struct ip_conntrack_helper sip[MAX_PORTS];


/* Not __exit: called from init() */
static void fini(void)
{
	int i;
	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		DEBUGP("ip_ct_sip: unregistering helper for port %d\n",
				ports[i]);
		ip_conntrack_helper_unregister(&sip[i]);
	}
}

static int __init init(void)
{
	int i, ret;
	int nRet = 0;
	//printk("ip_conntrack_sip v0.01 loading\n");
#if defined(VOICE_QUALITY_CONTROL_BY_ATM)	
#if defined(SUPPORT_RVSIP)
	for (i = 7; i <= NUM_PRI_PKT_GROUPS; i++){
#else	
	for (i = 2; i <= NUM_PRI_PKT_GROUPS; i++){
#endif	
		BcmAtm_SetPriorityPacketGroup(i,NULL,0);
	}
	stunaddr=(stunaddr_first<<16) | (stunaddr_end);
	stunmask=(stunmask_first<<16) | (stunmask_end);
#endif
	if (ports[0] == 0)
		ports[0] = SIP_PORT;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		memset(&sip[i], 0, sizeof(struct ip_conntrack_helper));
		sip[i].tuple.src.u.udp.port = htons(ports[i]);
		sip[i].tuple.dst.protonum = IPPROTO_UDP;
		sip[i].mask.src.u.udp.port = 0xFFFF;
		sip[i].mask.dst.protonum = 0xFFFF;
		sip[i].help = help;
		//sip[i].timeout = RTP_TIMEOUT;
		sip[i].timeout = 86400;
		DEBUGP("ip_ct_sip: registering helper for port %d\n", 
				ports[i]);
		ret = ip_conntrack_helper_register(&sip[i]);

		if (ret) {
			fini();
			return ret;
		}
		ports_c++;
	}
	return 0;
}


EXPORT_SYMBOL(ip_sip_lock);
EXPORT_SYMBOL(ip_conntrack_sip);
EXPORT_SYMBOL(find_sdp_rtp_addr);
EXPORT_SYMBOL(find_sdp_audio_port);
EXPORT_SYMBOL(find_sdp_video_port);
EXPORT_SYMBOL(find_sdp_image_port);
EXPORT_SYMBOL(expect_find);


module_init(init);
module_exit(fini);
