/* dhcpd.c
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"

//#define MY_DEBUG
/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config;

ODM_IP_MAP	*leaseMap;

/* Exit and cleanup */
static void exit_server(int retval)
{
	pidfile_delete(server_config.pidfile);
	CLOSE_LOG();
	exit(retval);
}


/* SIGTERM handler */
static void udhcpd_killed(int sig)
{
	sig = 0;
	LOG(LOG_INFO, "Received SIGTERM");
	exit_server(0);
}

/* Odm added function, used to check if the requested IP is in server lease range*/
int in_IPrange(u_int32_t ip) 
{	
	return   (ntohl(ip) >= ntohl(server_config.start) &&
			ntohl(ip) <= ntohl(server_config.end));
}

int myDEBUG(char *msg)
{
#ifdef MY_DEBUG
	printf("\n%s\n",msg);
#endif
	return 0;
}

#ifdef COMBINED_BINARY	
int udhcpd(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{	
	fd_set rfds;
	struct timeval tv;
        //BRCM --initialize server_socket to -1
	int server_socket = -1;
	int bytes, retval;
	struct dhcpMessage packet;
	unsigned char *state;
	char *server_id, *requested, *hostname;
	u_int8_t hostname_align[64];
	u_int32_t server_id_align, requested_align;
	unsigned long timeout_end;
	struct option_set *option;
	struct dhcpOfferedAddr *lease;
	struct sockaddr_in *sin;
	int pid_fd;
	char blank_chaddr[] = {[0 ... 15] = 0};
#ifdef AUTO_PROVISIONING_ON_LAN
       char verderId[256];
       unsigned char *optionPtr=NULL,*optionLenPtr=NULL;
       unsigned char optionLength;
#endif

	/* server ip addr */
	int fd = -1;
	struct ifreq ifr;

	argc = argv[0][0]; /* get rid of some warnings */
	
	OPEN_LOG("udhcpd");
	LOG(LOG_INFO, "udhcp server (v%s) started", VERSION);
	myDEBUG("DHCP server started");
	
	pid_fd = pidfile_acquire(server_config.pidfile);
	pidfile_write_release(pid_fd);

	memset(&server_config, 0, sizeof(struct server_config_t));
	
	read_config(DHCPD_CONF_FILE);

	//Odm Charles, load MAC-IP reservation mapping entries to leaseMap[]
	leaseMap = malloc(sizeof(ODM_IP_MAP)*DHCP_MAX_LEASE_MAP);
	memset(leaseMap,0,sizeof(ODM_IP_MAP)*DHCP_MAX_LEASE_MAP);
	read_leasemap(DHCP_LEASE_MAP_FILE);
	
	if ((option = find_option(server_config.options, DHCP_LEASE_TIME))) {
		memcpy(&server_config.lease, option->data + 2, 4);
		server_config.lease = ntohl(server_config.lease);
	}
	else server_config.lease = LEASE_TIME;	//default setting
	
	leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	//read_leases(server_config.lease_file);

	if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
		ifr.ifr_addr.sa_family = AF_INET;
		//Michael: fix brige mode enable dhcp cause error  strcpy(ifr.ifr_name, server_config.interface);
		strcpy(ifr.ifr_name, "br0");
		//strcpy(ifr.ifr_name, server_config.interface);
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
			sin = (struct sockaddr_in *) &ifr.ifr_addr;
			server_config.server = sin->sin_addr.s_addr;
			//Charles 07/18/2003, add server IP to lease table to prevent router giving away its own IP
			add_lease(blank_chaddr, server_config.server, 0x1fffffff, "DSLRouter");	//lease time=17 years!
			DEBUG(LOG_INFO, "%s (server_ip) = %s", ifr.ifr_name, inet_ntoa(sin->sin_addr));
		} else {
			LOG(LOG_ERR, "SIOCGIFADDR failed!");
			printf("###DHCPD Error--SIOCGIFADDR failed!\n");
			exit_server(1);
		}
		if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) {
			DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex);
			server_config.ifindex = ifr.ifr_ifindex;
		} else {
			LOG(LOG_ERR, "SIOCGIFINDEX failed!");
			printf("###DHCPD Error--SIOCGIFINDEX failed!\n");
			exit_server(1);
		}
		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
			memcpy(server_config.arp, ifr.ifr_hwaddr.sa_data, 6);
			DEBUG(LOG_INFO, "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x",
				server_config.arp[0], server_config.arp[1], server_config.arp[2], 
				server_config.arp[3], server_config.arp[4], server_config.arp[5]);
		} else {
			LOG(LOG_ERR, "SIOCGIFHWADDR failed!");
			printf("###DHCPD Error--SIOCGIFHWADDR failed!\n");
			exit_server(1);
		}
	} else {
		LOG(LOG_ERR, "socket failed!");
			printf("###DHCPD Error--socket failed!\n");
		exit_server(1);
	}
	close(fd);

#ifndef DEBUGGING
	pid_fd = pidfile_acquire(server_config.pidfile); /* hold lock during fork. */
	switch(fork()) {
	case -1:
		perror("fork");
		exit_server(1);
		/*NOTREACHED*/
	case 0:
		break; /* child continues */
	default:
		exit(0); /* parent exits */
		/*NOTREACHED*/
		}
	close(0);
	setsid();
	pidfile_write_release(pid_fd);
#endif


	signal(SIGUSR1, write_leases);
	signal(SIGTERM, udhcpd_killed);

	timeout_end = time(0) + server_config.auto_time;
	while(1) { /* loop until universe collapses */

		//Michael: fix brige mode enable dhcp cause error  
                //server_socket = listen_socket(INADDR_ANY, SERVER_PORT, server_config.interface);
                server_socket = listen_socket(INADDR_ANY, SERVER_PORT, "br0");
		if(server_socket == -1) {
			LOG(LOG_ERR, "couldn't create server socket -- au revoir");
			printf("###DHCPD Error--couldn't create server socket -- au revoir\n");
			exit_server(0);
		}			

		FD_ZERO(&rfds);
		FD_SET(server_socket, &rfds);
		if (server_config.auto_time) {
			tv.tv_sec = timeout_end - time(0);
			if (tv.tv_sec <= 0) {
				tv.tv_sec = server_config.auto_time;
				timeout_end = time(0) + server_config.auto_time;
				write_leases(0);
			}
			tv.tv_usec = 0;
		}
		retval = select(server_socket + 1, &rfds, NULL, NULL, server_config.auto_time ? &tv : NULL);
		if (retval == 0) {
			write_leases(0);
			timeout_end = time(0) + server_config.auto_time;
			close(server_socket);
                        //BRCM
                        server_socket = -1;
			continue;
		} else if (retval < 0) {
			DEBUG(LOG_INFO, "error on select");
			close(server_socket);
                        //BRCM
                        server_socket = -1;
			continue;
		}
		
		bytes = get_packet(&packet, server_socket); /* this waits for a packet - idle */
                //BRCM_BEGIN
		myDEBUG("DHCPD--got a packet!!");
		close(server_socket);
		if(bytes < 0)
			continue;

		if((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
			DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
			continue;
		}
		
#ifdef AUTO_PROVISIONING_ON_LAN
              if(state[0] == DHCPDISCOVER){
                  //printf("WILSON DEBUG: got DHCP discover message.\n");
                  if(optionPtr = get_option(&packet, DHCP_VENDOR)){
                      //printf("WILSON DEBUG: got DHCP VENDER option.\n");
                      optionLenPtr = optionPtr -1;
                      optionLength = optionLenPtr[0];
                      if(optionLength == strlen(ODM_VENDER_ID)){
                          strcpy(verderId, "");
                          memcpy(&verderId, optionPtr, optionLength);
                          verderId[optionLength]='\0';
                          if(strcmp(verderId, ODM_VENDER_ID) == 0){
                              //printf("WILSON DEBUG: got special verder identifier = [%s]\n", verderId);
                              continue;
                          }
                      }
                  }
              }
#endif

		lease = find_lease_by_chaddr(packet.chaddr);
		switch (state[0]) {
		case DHCPDISCOVER:
			DEBUG(LOG_INFO,"received DISCOVER");
			myDEBUG("DHCP server received DISCOVER");
			
			if (sendOffer(&packet) < 0) {
				LOG(LOG_ERR, "send OFFER failed -- ignoring");
				myDEBUG("DHCP server send OFFER failed!");
			}
			break;			
 		case DHCPREQUEST:
			DEBUG(LOG_INFO,"received REQUEST");
			myDEBUG("DHCP server received REQUEST");

			requested = get_option(&packet, DHCP_REQUESTED_IP);
			server_id = get_option(&packet, DHCP_SERVER_ID);
			//hostname = get_option(&packet, DHCP_HOST_NAME);
			//Odm added 05/27/2003 by Charles Chiu
			hostname = get_hostname(&packet);
			if(!hostname)
                              // Michael(10/31/2003): Fix Linux dhcp client cause dhcpdaemon disappear
                                   strcpy(&hostname_align,"Unknown");
		              //  strcpy(hostname,"Unknown");
			else strcpy(&hostname_align,hostname);
			
			//End Odm added

			if (requested) memcpy(&requested_align, requested, 4);
			if (server_id) memcpy(&server_id_align, server_id, 4);
		
			if (lease) {
				if (server_id) {
					/* SELECTING State */
					DEBUG(LOG_INFO, "server_id = %08x", ntohl(server_id_align));
					if (server_id_align == server_config.server && requested && 
					    requested_align == lease->yiaddr) {
						sendACK(&packet, lease->yiaddr);
					}
				} else {
					if (requested) {
						/* INIT-REBOOT State */
						if (lease->yiaddr == requested_align)
							sendACK(&packet, lease->yiaddr);
						else sendNAK(&packet);
					} else {
						/* RENEWING or REBINDING State */
						if (lease->yiaddr == packet.ciaddr)
							sendACK(&packet, lease->yiaddr);
						else {
							/* don't know what to do!!!! */
							sendNAK(&packet);
						}
					}						
				}
/*mask by wilson
				if (hostname) {
					bytes = hostname[-1];
					if (bytes >= (int) sizeof(lease->hostname))
						bytes = sizeof(lease->hostname) - 1;
					strncpy(lease->hostname, hostname, bytes);
					lease->hostname[bytes] = '\0';
				} else
					lease->hostname[0] = '\0';
*/
			} /* else remain silent --original udhcp design(RFC2131 Compliant)*/	
			//Odm modified code 2003/05/22 by Charles Chiu, 
			//ACK or NAK to unkown lease REQUESTs rather than keep silient.
			else {
				u_int32_t client_req_ip;
				if((packet.ciaddr==0) && requested)
					client_req_ip=requested_align;
				else
					client_req_ip=packet.ciaddr;
				
				if((in_IPrange(client_req_ip)) && !isReserved(client_req_ip) && 
					(!(lease = find_lease_by_yiaddr(client_req_ip)))) {					
					//IP in our range and is not occupied now. Add lease to lease repository
					//Charles 07/09/2003, AND NOT reserved (if it's the owner of this reserved IP, 
					//then it will be NAKed and go through DISCOVER again.)
					if (!add_lease(packet.chaddr, packet.ciaddr, server_config.offer_time,&hostname_align)) {
						LOG(LOG_WARNING, "lease pool is full -- OFFER abandoned");
						return -1;
					}
					sendACK(&packet, packet.ciaddr);
				}
				else
					sendNAK(&packet);
			}
			//End Odm modified code
			break;
		case DHCPDECLINE:
			DEBUG(LOG_INFO,"received DECLINE");
			myDEBUG("DHCP server received DECLINE");
			if (lease) {
				memset(lease->chaddr, 0, 16);
				lease->expires = time(0) + server_config.decline_time;
				//Odm added code by Charles Chiu 05/26/2003
				write_leases(0);
			}			
			break;
		case DHCPRELEASE:
			DEBUG(LOG_INFO,"received RELEASE");
			myDEBUG("DHCP server received RELEASE");
			if (lease) {
			lease->expires = time(0);
			//Odm added code by Charles Chiu 05/26/2003
			write_leases(0);
#ifdef CONFIG_INTOTO_IGATEWAY
			struct in_addr temp;
			temp.s_addr=lease->yiaddr;
			IGWEDnsDeleteStaticEntry(inet_ntoa(temp));
#endif
			}
			break;
		case DHCPINFORM:
			DEBUG(LOG_INFO,"received INFORM");
			myDEBUG("DHCP server received INFORM");
			send_inform(&packet);
			break;	
		default:
			LOG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring", state[0]);
			printf("DHCPD---got unsupported DHCP message (%02x) -- ignoring\n", state[0]);
		}
	}
        //BRCM_BEGIN 
        if (server_socket > 0)
          close(server_socket);
	return 0;
}

