/* 
 * leases.c -- tools to manage DHCP leases 
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"
#include "arpping.h"


/* clear every lease out that chaddr OR yiaddr matches and is nonzero */
void clear_lease(u_int8_t *chaddr, u_int32_t yiaddr)
{
	unsigned int i, blank_chaddr = 0, blank_yiaddr = 0;
	
	for (i = 0; i < 16 && !chaddr[i]; i++);
	if (i == 16) blank_chaddr = 1;
	blank_yiaddr = (yiaddr == 0);
	
	for (i = 0; i < server_config.max_leases; i++)
		if ((!blank_chaddr && !memcmp(leases[i].chaddr, chaddr, 16)) ||
		    (!blank_yiaddr && leases[i].yiaddr == yiaddr)) {
			memset(&(leases[i]), 0, sizeof(struct dhcpOfferedAddr));
			//Charles 08/28/2003, delete this entry from Embedded DNS db
#ifdef CONFIG_INTOTO_IGATEWAY
			struct in_addr temp;
			temp.s_addr=yiaddr;
			IGWEDnsDeleteStaticEntry(inet_ntoa(temp));
#endif
			/*Odm added code 05/26/2003 by Charles Chiu*/
			write_leases(0);
		}
}


/********************************************************************************* 
//add a lease into the table, clearing out any old ones 
// Odm Charles modified 06/05/2003---
// "hostname" assign rule: if hostname is assigned in MAC-IP reservation table (leaseMap), use it;
//					     otherwise use the provided hostname.
*********************************************************************************/
struct dhcpOfferedAddr *add_lease(u_int8_t *chaddr, u_int32_t yiaddr, unsigned long lease, char *hostname)
{
	struct dhcpOfferedAddr *oldest;
	char blank_chaddr[] = {[0 ... 15] = 0};
	
	//look up hostname in leaseMap
	int i, found=0;
	for(i=0;i<DHCP_MAX_LEASE_MAP && !found;i++)
		if(!memcmp(leaseMap[i].mac,chaddr,6)) {
			found=1;
			//check for NULL hostname assignment, Charles 07/02/2003
			if(strlen(leaseMap[i].hname)>0)
				memcpy(hostname, leaseMap[i].hname,64);
		}
	
	/* clean out any old ones */
	clear_lease(chaddr, yiaddr);
		
	oldest = oldest_expired_lease();
	
	if (oldest) {
		memcpy(oldest->chaddr, chaddr, 16);
		oldest->yiaddr = yiaddr;
		oldest->expires = time(0) + lease;
		memcpy(oldest->hostname, hostname, 64);
		//Charles 08/28/2003, add this entry into Embedded DNS db
#ifdef CONFIG_INTOTO_IGATEWAY
		if(strcmp(hostname, "DSLRouter") != 0 && 
			memcmp(blank_chaddr,chaddr, 6) != 0  &&
			strcmp(hostname, "" ) != 0  && yiaddr > 0 ) 
			{
			struct in_addr temp;
			temp.s_addr=yiaddr;
			IGWEDnsAddStaticEntry(inet_ntoa(temp), hostname);
		}
#endif
		/*Odm added code 05/26/2003 by Charles Chiu*/
		write_leases(0);
	}
	
	return oldest;
}


/* true if a lease has expired */
int lease_expired(struct dhcpOfferedAddr *lease)
{
	return (lease->expires < (unsigned long) time(0));
}	


/* Find the oldest expired lease, NULL if there are no expired leases */
struct dhcpOfferedAddr *oldest_expired_lease(void)
{
	struct dhcpOfferedAddr *oldest = NULL;
	unsigned long oldest_lease = time(0);
	unsigned int i;

	
	for (i = 0; i < server_config.max_leases; i++)
		if (oldest_lease > leases[i].expires) {
			oldest_lease = leases[i].expires;
			oldest = &(leases[i]);
		}
	return oldest;
		
}

//Odm DEBUG
void print_mac(u_int8_t *chaddr)
{
		u_int8_t mac[6];
		int i;

		memcpy(&mac,chaddr,6);
		for (i = 0; i < 6; i++) {
			printf("%02x", mac[i]);
			if (i != 5) printf(":");
		}
		printf("\n");
}

/* Find the first lease that matches chaddr, NULL if no match 
// Odm Charles 06/05/2003 modified,
// look up leaseMap to see if there is a reserved IP for this MAC,
// If YES, add a entry to lease table and return it
*/
struct dhcpOfferedAddr *find_lease_by_chaddr(u_int8_t *chaddr)
{
	unsigned int i, found=0;

	//Check the curent lease table FIRST!!!
	for (i = 0; i < server_config.max_leases; i++)
		if (!memcmp(leases[i].chaddr, chaddr, 16)) {
			//printf("DHCP--find_lease_by_chaddr FOUND, lease[%d].chaddr=",i);
			//print_mac(leases[i].chaddr);
			//printf("DHCP--find_lease_by_chaddr FOUND, chaddr=");
			//print_mac(chaddr);
			return &(leases[i]);
		}

	//Then look up leaseMap (IP reservation table)
	for(i=0;i<DHCP_MAX_LEASE_MAP && !found;i++)
		if(!memcmp(leaseMap[i].mac,chaddr,6)) {
			//printf("DHCP--find_lease_by_chaddr FOUND, leaseMap[%d].mac=",i);
			//print_mac(leaseMap[i].mac);
			found=1;
			//if IP is leagal (in range) and available, add an entry to lease table
			if(	(in_IPrange(leaseMap[i].ip)) && 
				(!(find_lease_by_yiaddr(leaseMap[i].ip)))) 
				return add_lease(	chaddr, leaseMap[i].ip, 
									time(0)+server_config.lease, 
									leaseMap[i].hname);
		}
		
	//printf("DHCP--find_lease_by_chaddr NOT FOUND!!-- chaddr=");
	//print_mac(chaddr);
	return NULL;
}


/* Find the first lease that matches yiaddr, NULL is no match */
struct dhcpOfferedAddr *find_lease_by_yiaddr(u_int32_t yiaddr)
{
	unsigned int i;
	struct in_addr temp;

	temp.s_addr=yiaddr;

	for (i = 0; i < server_config.max_leases; i++)
		if (leases[i].yiaddr == yiaddr) {
			//printf("DHCP--find_lease_by_yiaddr %s found!\n",inet_ntoa(temp));
			return &(leases[i]);
		}
	//printf("DHCP--find_lease_by_yiaddr %s NOT found!\n",inet_ntoa(temp));
	return NULL;
}

//Odm Charles, check if this IP is a reserved IP (on MAC-IP mapping table)
int isReserved(u_int32_t ip) {
	int i, found=0;

	if(ip==0) return -1;

	for(i=0;i<DHCP_MAX_LEASE_MAP && !found;i++)
		if(leaseMap[i].ip==ip) {
			found=1;
			struct in_addr temp;
			temp.s_addr=ip;
			//printf("DHCP--found reserved IP---%s\n",inet_ntoa(temp));
		}

	return found;
}

/* find an assignable address, it check_expired is true, we check all the expired leases as well.
 * Maybe this should try expired leases by age... */
u_int32_t find_address(int check_expired) 
{
	u_int32_t addr, ret = 0;
	struct dhcpOfferedAddr *lease = NULL;
	struct in_addr temp1, temp2;

	addr = server_config.start;
	temp1.s_addr=addr;
	temp2.s_addr=server_config.end;
	//printf("DHCP--find_address: start=%s, end=%s\n",inet_ntoa(temp1),inet_ntoa(temp2));
	// brcm
	for (;ntohl(addr) <= ntohl(server_config.end) ;addr = htonl(ntohl(addr) + 1)) {
		//Odm DEBUG
			temp1.s_addr=addr;
			//printf("DHCP--find_address is trying this IP--%s\n",inet_ntoa(temp1));
		//end	

		/* ie, 192.168.55.0 */
		if (!(ntohl(addr) & 0xFF)) continue;

		/* ie, 192.168.55.255 */
		if ((ntohl(addr) & 0xFF) == 0xFF) continue;

		//Odm Charles, check for reserved IP
		//won't use reserved IP unless no other free IP available
		if(!check_expired && isReserved(addr))	continue;

		/* lease is not taken */
		if ((!(lease = find_lease_by_yiaddr(addr)) ||

		     /* or it expired and we are checking for expired leases */
		     (check_expired  && lease_expired(lease))) &&

		     /* and it isn't on the network */
	    	     !check_ip(addr)) {
			ret = addr;
			temp1.s_addr=addr;
			//printf("DHCP--find_address found free IP!--%s\n",inet_ntoa(temp1));
			break;
		}
	}
	temp1.s_addr=ret;
	//printf("DHCP--find_address Returned IP=%s\n",inet_ntoa(temp1));
	return ret;
}


/* check is an IP is taken, if it is, add it to the lease table */
int check_ip(u_int32_t addr)
{
	char blank_chaddr[] = {[0 ... 15] = 0};
	char blank_hostname[] = {[0 ... 31] = 0};
	struct in_addr temp;
	//printf("DHCP--check_ip start...\n");
	if (!arpping(addr, server_config.server, server_config.arp, "br0")) {
		temp.s_addr = addr;
	 	LOG(LOG_INFO, "%s belongs to someone, reserving it for %ld seconds", 
	 		inet_ntoa(temp), server_config.conflict_time);
	 	//printf("DHCP--check_ip %s belongs to someone, add it to lease table\n",inet_ntoa(temp));
		add_lease(blank_chaddr, addr, server_config.conflict_time, blank_hostname);
		return 1;
	} else {
		//temp.s_addr=addr;
		//printf("DHCP--check_ip for IP %s AVAILABLE!\n",inet_ntoa(temp));
		return 0;
	}
}

/****************************************************************
 * Charles 08/28/2003, Copy from conf45/util/igateway/igwwrap.c
 ****************************************************************/
#define DNSRD_AGENT_NAME "/var/DnsrdAgent"
void IGWEDnsAddStaticEntry(char *IpAddr,char *DomainName)
{
  int AgentSockFd, len;
  struct sockaddr_un AgentSockAddr;
  char buff[255]="";

//printf("###DHCPD IGWEDnsAddStaticEntry called, IP=%s, hostname=%s\n", IpAddr, DomainName);

	AgentSockFd = socket(AF_UNIX, SOCK_STREAM, 0);
	if(AgentSockFd < 0) {
		printf("###DHCPD IGWEDnsAddStaticEntry--Can't not init socket\n");
		return;
	}

	AgentSockAddr.sun_family = AF_UNIX;
	strcpy(AgentSockAddr.sun_path, DNSRD_AGENT_NAME);
	if(connect(AgentSockFd, (struct sockaddr *) &AgentSockAddr, sizeof(struct sockaddr_un)) < 0) {
		close(AgentSockFd);
		printf("###DHCPD IGWEDnsAddStaticEntry--connect Fail\n");
		return;
	}
	/* 4 for Adding Static Entry */
  	sprintf(buff, "%c%4d%s%s%4d%s%s", 4,strlen(IpAddr),";",IpAddr,strlen(DomainName),";",DomainName);
	len = strlen(buff);
	write(AgentSockFd, buff, len);
	close(AgentSockFd);
}

void IGWEDnsDeleteStaticEntry(char *IpAddr)
{
  int AgentSockFd, len;
  struct sockaddr_un AgentSockAddr;
  char buff[255]="";

//printf("###DHCPD IGWEDnsDeleteStaticEntry called, ip=%s\n", IpAddr);

  AgentSockFd = socket(AF_UNIX, SOCK_STREAM, 0);
  if(AgentSockFd < 0) {
    printf("###DHCPD IGWEDnsDeleteStaticEntry--Can't not init socket\n");
    return;
  }

  AgentSockAddr.sun_family = AF_UNIX;
  strcpy(AgentSockAddr.sun_path, DNSRD_AGENT_NAME);
  if(connect(AgentSockFd, (struct sockaddr *) &AgentSockAddr, sizeof(struct sockaddr_un)) < 0) {
    close(AgentSockFd);
    printf("###DHCPD IGWEDnsDeleteStaticEntry--connect Fail\n");
    return;
  }
  /* 5 for Deleting Static Entry */
  sprintf(buff, "%c%4d%s%s", 5,strlen(IpAddr),";",IpAddr);
  len = strlen(buff);
  write(AgentSockFd, buff, len);
  close(AgentSockFd);
}

